#ifndef AUX_HH
#define AUX_HH

////////////////////////////////////////////////////////////
//
// classes Index, Sequence, Structure, and Alignment
//
// auxilliary classes for LSSAligner
//
// (C) Sebastian Will, 2003
//
////////////////////////////////////////////////////////////


#include <stack>
#include <list>
#include<algorithm>

typedef pair<int,int> arc_t;

class Index {
public:
  Index(): i0(-1), i1(-1), q(-1) {}
  Index(int i0_, int i1_) : i0(i0_), i1(i1_), q(-1) {}
  Index(int i0_, int i1_, int q_) : i0(i0_), i1(i1_), q(q_) {}

  void set(int i0_, int i1_, int q_) {
    i0 = i0_;
    i1 = i1_;
    q  = q_;
  }

  int i0;
  int i1;
  int q;
};


// ------------------------------------------------------------
// The Sequence class stores and gives access to a sequence string
//
// The first index of a sequence is 1 (instead of 0 as in a C-string)!!!
// (This is achieved by adding '-' in front of the string)
//
class Sequence : private string {
public:
  Sequence(): string(),len(0) {};
  Sequence(char *s): string(upcase("-"+string(s))), len(string::length()-1) {}
  Sequence(const string &s): string(upcase("-"+s)), len(string::length()-1) {}
  // virtual ~Sequence() {}
  
  char operator [] (int i) const {
    assert(1<=i); assert(i<=len);
    return string::operator [](i);
  }
  int length() const {return len;}

  const Sequence &operator =(string s) {
    string::operator =(upcase("-"+s));
    len=string::length()-1;
    return *this;
  }

  string upcase(const string &s) const {
    string t;
    t.resize(s.length());
    for (unsigned int i=0; i<s.length(); i++)
      t[i]=toupper(s[i]);
    return t;
  }

  friend ostream & operator <<(ostream &out, const Sequence &s);
private: 
  int len;
};


// ------------------------------------------------------------
//
// Class for representing a structure
// 
// A structure is a vector of arcs (i.e. pairs of positions in the sequence)
//
// A structure may be constructed by a bracket string describing the structure
// or empty.
//
// The class maintains several arrays to answer queries in constant time.
// => space requirement is O(sequence length)
// ( answering arc-complete query from O(seq len) arrays in constant time
//   works by comparing the index of the enclosing arc of both ends.
//   This is not trivial, since left and right arc ends need special handling!)
//
//

class Structure : private vector<arc_t> {
public:

  Structure(int seqlength=0) : seqlen(seqlength) {
    init_all();
  }  
  Structure(string s) {
    *this = s;
  }
  
  const Structure& operator =(string s) {
    seqlen=s.length();
    init_all();

    // The arcs are pushed in the order first inner then outer arcs.
    //
    stack<int> st;
    for (int i=seqlen-1; i>=0; i--)
      if (s[i]==')') st.push(i);
      else if (s[i]=='(') {
	push_back(arc_t(i+1,st.top()+1));
	st.pop();
      }
    structure="-"+s+"-";
    return *this;
  }

  // virtual ~Structure() {}

  arc_t operator [] (int i) const {return vector<arc_t>::operator [](i);}

  int size() const {return vector<arc_t>::size();} 
    
  bool inc(int pos) const {
    return adj_vec[pos]>0;
  }
  
  int adj(int pos) const {
    return adj_vec[pos];
  }
  
  // bool is_arc(int i,int j) const {return adj(i)==j;}
  
  bool end_left(int k) const {
    return adj(k)>k;
  }
  bool end_right(int k) const {
    return adj(k)!=0 && adj(k)<k;
  }

  // bool adj_in(int k, int i, int j) const {
  //  return inc(k) && i<=adj(k) && adj(k)<=j;
  // }

  int index(int i) const {
    return idx_vec[i];
  };

  string getString() const {return structure;}
  
  char toChar(int i) const {
    if (!inc(i)) return '.';
    if (adj(i)>i) return '('; else return ')';
  }
  int seqlength() const {return seqlen;}

  bool test_arc_compl(int i, int j) const {
    int count=0;
    
    for (int k=i; count>=0 && k<=j; k++) {
      switch (structure[k]) {
      case '(':
	count++; break;
      case ')':
	count--; break;
      }
    }
    return (count==0);
  }
  
  bool arc_compl(int i, int j) const {
    assert(i<=j);
    assert(i>=1); assert(j<=seqlen);
        
    bool res = left_encl[i] == right_encl[j];

#ifndef NDEBUG
    if (test_arccompl(i,j) != res) {
      cerr << "arc_compl: [count:" <<count<<" res:"<<res<<"] "<< structure.substr(i,j-i+1)<<endl;
      cerr << "i:"<<i<<" j:"<<j<<" left_encl[i]:"<<left_encl[i]<<" right_encl[j]:"<<right_encl[j]<<endl;
      exit(-1);
    }
#endif
    
    return res;
  }

  friend ostream & operator <<(ostream &out, const Structure &p);

private:

  // push backs must be in the order inner to outer arcs
  void push_back(const arc_t &a) {
    for (int i=a.first; i<=a.second; i++) {
      if (i!=a.first && left_encl[i]==-1) left_encl[i]=size();
      if (i!=a.second && right_encl[i]==-1) right_encl[i]=size();
    }

    idx_vec[a.first]=idx_vec[a.second]=size();
    
    vector<arc_t>::push_back(a);
    adj_vec[a.first] = a.second;
    adj_vec[a.second] = a.first;
  }

  void init_all() {
    adj_vec.resize(seqlen+1);
    idx_vec.resize(seqlen+1);
    left_encl.resize(seqlen+1);
    right_encl.resize(seqlen+1);

    for (int i=0; i<=seqlen; i++) {
      adj_vec[i]=0;
      idx_vec[i]=-1;
      left_encl[i]=-1;
      right_encl[i]=-1;
    }
  }

  vector<int> adj_vec; // for arc ends, other end of arc, otherwise 0
  vector<int> idx_vec; // for arc ends, index of arc, otherwise -1

  vector<int> left_encl; // index of enclosing arc, the left end of an arc >a< is not enclosed by >a<, while right ends are
  vector<int> right_encl; // index of enclosing arc, right ends of arcs are not enclosed

  int seqlen;

  string structure;
};

inline
ostream & operator <<(ostream &out, const Structure &p) {
  for (int i=1; i<=p.seqlength(); i++)
    out << p.toChar(i);  
  
  return out;
}

inline
ostream & operator <<(ostream &out, const Sequence &s) {
  for (int i=1; i<=s.length(); i++) 
    out << s[i];
  return out;
}

// ein Alignment ist eine Menge von Paaren
// der aufeinander alignierten Positionen
class Alignment :  public vector<pair<int,int> > {
public:
  Alignment(): vector<pair<int,int> >() {}

  Alignment & append(int i0, int i1) {
    push_back(pair<int,int>(i0,i1));
    return *this;
  }

  friend ostream &operator <<(ostream &out, const Alignment &a);
  
  bool is_edge(int i0,int i1) const {
    return find( begin(), end(), pair<int,int>(i0,i1) ) != end();
  }
};

inline
ostream & operator <<(ostream &out, const Alignment &a) {
  for (Alignment::const_iterator i=a.begin(); i != a.end(); ++i) {
    out << i->first << " " << i->second <<endl;
  }
  return out;
}

#endif // AUX_HH
