////////////////////////////////////////////////////////////
//
// class LSSAligner
//
// performs local sequence structure alignment
// computes similarity score and traceback
//
// (C) Sebastian Will, 2003
//
////////////////////////////////////////////////////////////

#include <vector>
#include <string>
#include "lssaligner.hh"


// minimal length of exclusions
// length <= 1 will allow any exclusions (without time overhead)
// length > sequence length will turn of exclusions, 
// thereby effectively forcing a global (however still sequence local) alignment
const int min_excl_len = 0;

/* --------------------------------------------------
//  index names:
//  
//   /-------------- a0 ---------------\
//   a0l        k0                 j0  a0r
//  -|----------|------------------|---|------- seq 0
//  ---|-------------|----------|---------|---- seq 1
//     a1l           k1         j1        a1r
//     \----------------- a1 -------------/
//  
//  q is a state, i.e. index of a submatrix
//
// -------------------------------------------------- */



// ===========================================================================
//
// Init and recursion cases
//
// ===========================================================================

// ------------------------------------------------------------
// method to compute one recursion case
// if the case is maximal, then set new maxscore and maxfrom
void new_maximum(const score_matrix_t &s, int from_i0, int from_i1, int from_q,
		 score_t newscoreadd, score_t *maxscore,  Index *maxfrom)
{
  score_t newscore = s.elem(from_i0, from_i1) + newscoreadd;
  
  if (newscore > *maxscore) {
    *maxscore = newscore;
    if (maxfrom != NULL)
      maxfrom->set(from_i0,from_i1,from_q);
  }
}

// ------------------------------------------------------------
// standard initialization for arc-match and top level
score_t LSSAligner::std_init_column(score_matrix_t &M, int j0, int a1l) {
  score_t score = M.elem(j0-1,a1l);
  
  if (p0.end_left(j0))
    score += sim.break_left(0,p0.index(j0),j0,-1);
  else if (p0.end_right(j0))
    score += sim.break_right(0,p0.index(j0),j0,-1);
  else
    score += sim.base(j0,-1);
  
  return score;
}

score_t LSSAligner::std_init_row(score_matrix_t &M, int a0l, int j1) {
  score_t score = M.elem(a0l,j1-1);
  
  if (p1.end_left(j1))
    score += sim.break_left(1,p1.index(j1),-1,j1);
  else if (p1.end_right(j1))
    score += sim.break_right(1,p1.index(j1),-1,j1);
  else
    score += sim.base(-1,j1);
  
  return score;
}

// ------------------------------------------------------------
// standard recursion cases in arc matches and cases for top level
inline score_t LSSAligner::std_rec_cases(score_t maxscore, const score_matrix_t &s, 
			    int a0l, int a1l, int j0, int j1,
			    int a0r, int a1r, Index *maxfrom=NULL) {
  // base match/ins/del can break arcs

  // base match
  bool free0=false;
  bool free1=false;
  score_t match_add=0;
  score_t ins_add=0;
  score_t del_add=0;

  if ( p0.end_left(j0) ) {
    match_add += sim.break_left(0,p0.index(j0),j0,j1);
    del_add   += sim.break_left(0,p0.index(j0),j0,-1);
  }
  else if ( p0.end_right(j0) ) {
    match_add += sim.break_right(0,p0.index(j0),j0,j1);
    del_add   += sim.break_right(0,p0.index(j0),j0,-1);
  } else { 
    del_add = sim.base(j0,-1);
    free0 = true;
  }

  if ( p1.end_left(j1) ) {
    match_add += sim.break_left(1,p1.index(j1),j0,j1);
    ins_add   += sim.break_left(1,p1.index(j1),-1,j1);
  }
  else if ( p1.end_right(j1) ) {
    match_add += sim.break_right(1,p1.index(j1),j0,j1);
    ins_add   += sim.break_right(1,p1.index(j1),-1,j1);
  } else {
    ins_add = sim.base(-1,j1);
    free1 = true;
  }

  if ( free0 && free1 ) 
    match_add = sim.base(j0,j1);
  
  new_maximum(s, j0-1, j1-1, 0, match_add, &maxscore, maxfrom);
  
  // base del
  new_maximum(s, j0-1, j1, 0, del_add, &maxscore, maxfrom);
  
  // base ins
  new_maximum(s, j0, j1-1, 0, ins_add, &maxscore, maxfrom);
  
  // arc match
  if ( p0.end_right(j0) && p1.end_right(j1) ) {
    new_maximum(s, p0.adj(j0)-1, p1.adj(j1)-1, 0,
		AM.elem(p0.index(j0),p1.index(j1)),
		&maxscore, maxfrom);
  }
  return maxscore;
}

// ------------------------------------------------------------
// special cases to handle locality
//
// computes best scores for j0,j1 in each of the four matrices M
// by admissible exclusions
//
// input parameters:
//  const score_matrix_t M[] --- the four M matrices, read-only parameter
//  score_t maxscore[]       --- maximal scores in each state, this is a input/output parameter,
//                               since special cases are handled after standard ones, we need input,
//                               this is computed by the function (optionally together with maxfrom for trace)
//  int a0l,int a1l          --- left end of the enclosing arc-match, indexes of the first validmatrix row/column
//  int j0, int j1           --- matrix cell to compute (best alignment of prefixes a0l+1,...,j0 and a1l+1,...,j1)
//  int a0r,int a1r          --- right end of the enclosing arc-match, UNUSED!!!
//  Index *maxfrom=NULL      --- only for traceback, ignored for filling of matrix (in score computation phase)
//
inline void LSSAligner::special_rec_cases(const score_matrix_t M[], score_t maxscore[], 
			     int a0l,int a1l, int j0, int j1,
			     int a0r, int a1r, Index *maxfrom=NULL) {
  Index *maxfromP[4];
  if (maxfrom==NULL) {
    maxfromP[0]=maxfromP[1]=maxfromP[2]=maxfromP[3]=NULL;
  } else {
    maxfromP[0]=&maxfrom[0]; maxfromP[1]=&maxfrom[1];
    maxfromP[2]=&maxfrom[2]; maxfromP[3]=&maxfrom[3];
  }

  // * int k0 --- left end of the exclusion
  // * if an exclusion in sequence 0 is allowed, the matrix elem j0,j1
  //   may be optimally derived from k0-1,j1 by skipping the exclusion [k0..j0]
  // * The largest possible k0 is j0 (there the single base j0 is excluded)
  // * The smallest possible k0 is a0l+1 (there the whole prefix in the enclosing arc-match is excluded)
  // * [k0..j0] has to be arc-complete
  //
  for (int k0=j0; k0>a0l; k0--) {
    // jump over arcs
    // and break if [k0..j0] cannot be arc-complete for smaller k0
    if (p0.end_right(k0)) {   // here, [k0..j0] cannot be ac, since k0 is the right end of an arc a0'
      k0 = p0.adj(k0);        // set k0 to the largest k0, where [k0..j0] is ac again,
                              // for this aim, k0 is set to the left end of arc a0'
                              // Note that we cannot jump outside of the enclosing arc, due to nestedness.
    } else if (p0.end_left(k0)) {   // if we get in any other way to a left end, break,
                                    // since then, there are more left ends than right ends.
                                    // This cannot be balanced by arc ends with smaller indices
      break;
    }
    // by the above, [k0..j0] is arc-complete
    
    if (min_excl_len<=1 || j0-k0>min_excl_len) {
      new_maximum(M[ALLOW_EXCL_IN_1], k0-1, j1, ALLOW_EXCL_IN_1, 0,
		  &maxscore[ALLOW_BOTH_EXCLS], maxfromP[ALLOW_BOTH_EXCLS]);
      
      new_maximum(M[ALLOW_NO_EXCLS], k0-1, j1, ALLOW_NO_EXCLS, 0,
		  &maxscore[ALLOW_EXCL_IN_0], maxfromP[ALLOW_EXCL_IN_0]);    
    }
  }

  // analogously for exclusion in second sequence/structure 
  for (int k1=j1; k1>a1l; k1--) {
    if (p1.end_right(k1)) {
      k1 = p1.adj(k1);
    } else if (p1.end_left(k1)) {
      break;
    }
    if (min_excl_len<=1 || j1-k1>min_excl_len) {
      new_maximum(M[ALLOW_EXCL_IN_0], j0, k1-1, ALLOW_EXCL_IN_0, 0,
		  &maxscore[ALLOW_BOTH_EXCLS], maxfromP[ALLOW_BOTH_EXCLS]);
      
      new_maximum(M[ALLOW_NO_EXCLS], j0, k1-1, ALLOW_NO_EXCLS, 0,
		  &maxscore[ALLOW_EXCL_IN_1], maxfromP[ALLOW_EXCL_IN_1]);
    }
  }
}


// ===========================================================================
//
// Computation of score -- filling of DP matrices
//
// ===========================================================================


// -------------------------------------------------------------------------
// compute the top level matrix for subsequences l0+1..r0-1 with l1+1..r1-1
// (i.e. compute a matrix with indices [l0..r0-1]x[l1..r1-1]
//
//
void LSSAligner::compute_top_level(int l0, int l1, int r0,int r1) {
  // initialize T  
  T.elem(l0,l1)=0;
  for (int j0=l0+1; j0<r0; j0++) {
    score_t maxscore = std_init_column(T,j0,l1); 
    if (p0.arc_compl(l0+1,j0)) maxscore=max((score_t)0,maxscore);
    T.elem(j0,l1) = maxscore;
  }
  for (int j1=l1+1; j1<r1; j1++) {
    score_t maxscore = std_init_row(T,l0,j1);
    if (p1.arc_compl(l1+1,j1)) maxscore=max((score_t)0,maxscore);
    T.elem(l0,j1) = maxscore;
  }
  // fill T
  for (int j0=l0+1; j0<r0; j0++)
    for (int j1=l1+1; j1<r1; j1++) {
      score_t minscore;
      if (p0.arc_compl(l0+1,j0)
	  && p1.arc_compl(l1+1,j1))
	minscore = 0;
      else  minscore = NEG_SCORE_INFTY;
      
      T.elem(j0,j1) = 
	std_rec_cases(minscore, T, l0, l1, j0, j1, r0, r1);
    }
}

// ------------------------------------------------------------
void LSSAligner::compute_M_matrices(int a0l,int a1l,int a0r,int a1r) {
  // compute M[q](j0,j1) in maxscore[q]
  score_t maxscore[4];

  // ------------------------------------------------------------
  // initialize submatrices M
 
  // termination
  for (int q=0; q<4; q++)
    M[q].elem(a0l,a1l) = 0;

  // init first column
  for (int j0=a0l+1; j0<a0r; j0++) {  
    for (int q=0; q<4; q++)
      maxscore[q] = std_init_column(M[q],j0,a1l);
    special_rec_cases(M,maxscore, a0l, a1l, j0, a1l,
		      a0r, a1r);
    for (int q=0; q<4; q++)
      M[q].elem(j0,a1l) = maxscore[q];
  }
  // init first line
  for (int j1=a1l+1; j1<a1r; j1++) {
    for (int q=0; q<4; q++)
      maxscore[q] = std_init_row(M[q],a0l,j1);
    special_rec_cases(M,maxscore, a0l, a1l, a0l, j1,
		      a0r, a1r);
    for (int q=0; q<4; q++)
      M[q].elem(a0l,j1) = maxscore[q];
  }
  // fill matrices according to recursion scheme      
  for (int j0=a0l+1; j0 < a0r; j0++) 
    for (int j1=a1l+1; j1 < a1r; j1++) {      
      // --------------------
      // first, handle the cases that are equal for all submatrices
      for (int q=0; q<4; q++)
	maxscore[q]=std_rec_cases(NEG_SCORE_INFTY,M[q],
				  a0l,a1l,j0,j1,a0r,a1r);
      // --------------------
      // then, handle special cases for locality
      special_rec_cases(M,maxscore, a0l, a1l, j0, j1, a0r, a1r);
      // --------------------
      // write matrices M
      for (int q=0; q<4; q++)
	M[q].elem(j0,j1) = maxscore[q];
    }
}

// ------------------------------------------------------------
score_t LSSAligner::score() {  
  // we use the submatrices M for every arc-match
  
  // ----------------------------------------
  // first, compute AM

  // loop over all O(n^2) many arc-pairs
  // in the nested structure p0 we go from inner arcs to outer ones
  // p0 is assumed to be sorted!
  for (int a0=0; a0<plen0 ; a0++) {
    if (VERBOSE) {cout <<"#"; cout.flush();}
    for (int a1=0; a1<plen1; a1++) {
      
      // compute AM(a0,a1) in O(n^3)
      
      // align from right to left while allowing 
      // maximally one exclusion in each sequence      
      //
      
      int a0l = p0[a0].first;
      int a1l = p1[a1].first;
      
      int a0r = p0[a0].second;
      int a1r = p1[a1].second;

      compute_M_matrices(a0l,a1l,a0r,a1r);

      /*
      // get maximal entry (a0r-1,a1r-1) in any submatrix:
      // the maximal entry is M[0].elem(a0r-1,a1r-1) !
      */
      
      AM.elem(a0,a1) = sim.arc_match(a0,a1) + M[0].elem(a0r-1,a1r-1);

      if (VERBOSE) {
	cout <<"Matrices for arc-pair ("<<a0<<", "<<a1<<"), "
	     <<"i.e. (("<<p0[a0].first<<", "<<p0[a0].second<<"), "
	     <<"("<<p1[a1].first<<", "<<p1[a1].second<<"))"
	     <<endl;
	for (int q=0; q<4; q++) {	  
	  cout << "Matrix M["<<q<<"]:"<<endl;
	  M[q].print(cout,a0l,a1l,a0r,a1r);
	}
      }

    }
  }
  // end loop over arc-pairs
  // --------------------

  // ----------------------------------------
  // then, compute the final score using AM
  // the match of the two sequences is different
  // from arc-matches regarding the handling of locality
  //

  if (VERBOSE) {
    cout <<"Matrix AM"<<endl;
    AM.print(cout);

    cout << endl << "top level"<<endl;
  }

  /* ----------------------------------------
    TOP LEVEL

    To get the best alignment of arc-complete subsequences,
    the first idea was to search over all starting points i0,i1.
    Since there is a dependency (arc-completeness) between
    the starts i0,i1 and ends j0,j1 of the subsequences,
    we cannot immediately apply the trick of local sequence alignment,
    in order to reduce complexity.

    However, a similar trick is possible to compute the best
    score for a whole set of starts i0,i1 at the same time.
    In local sequence alignment arbitrary prefixes can be dropped,
    we allow to drop arc-complete prefixes. (Crucial property:
    if the whole sequences are arc-complete, the subsequences with
    dropped arc-complete prefix are arc-complete.)
    Thus the starts computed at the same time are in fact all starts at
    the same level in the arc-hierarchy.

    Complexity is reduced from O(n^2 m^2) to O(|P1||P2|nm)
  */

  // cout << "Top level"<<endl;

  score_t maxscore=NEG_SCORE_INFTY;
  max_l0=1; max_l1=0;
  max_i0=1; max_i1=0;
  
  for (int a0=0; a0<=plen0 ; a0++) {
    int l0; int r0;
    if (a0!=plen0) {l0=p0[a0].first; r0=p0[a0].second;}
    else {l0=0; r0=slen0+1;}
    
    for (int a1=0; a1<=plen1; a1++) {
      int l1; int r1;
      if (a1!=plen1) {l1=p1[a1].first; r1=p1[a1].second;}
      else {l1=0; r1=slen1+1;}

      compute_top_level(l0,l1,r0,r1);

      // search the maximal entry in the top level matrix, where
      // the corresponding sequences are arc-complete 
      // POSSIBLE OPTIMIZATION: jump over arcs (small effect)
      for (int i0=l0+1; i0<r0; i0++) {
	if (p0.arc_compl(l0+1,i0))
	  for (int i1=l1+1; i1<r1; i1++) {
	    if (p1.arc_compl(l1+1,i1))
	      if (T.elem(i0,i1) > maxscore) {
		max_l0=l0; max_l1=l1;
		max_i0=i0; max_i1=i1;
		maxscore = T.elem(i0,i1);
	      }
	  }
      }
    }
    if (VERBOSE) {cout <<"#"; cout.flush();}
  }
  // cout << max_l0<<" "<< max_l1<<" "<< max_i0 <<" "<<  max_i1<<endl;
  
  return maxscore;
}


// ===========================================================================
//
// Traceback
//
// ===========================================================================


void LSSAligner::trace_M_matrices(int a0l, int a1l, int a0r, int a1r,
			    int q, int j0, int j1) {
  if (j0>a0l && j1>a1l) {
    Index idxs[4];
    score_t maxscore[4];
  
    // --------------------
    // init maxscore
    for (int q_=0; q_<4; q_++)
      maxscore[q_]=NEG_SCORE_INFTY;
    
    // --------------------
    // first, handle the cases that are equal for all submatrices
    for (int q_=0; q_<4; q_++) {	    
      maxscore[q_]=std_rec_cases(maxscore[q_],M[q_],
				 a0l,a1l,j0,j1,a0r,a1r,&idxs[q_]);
      idxs[q_].q=q_;
    }
    
    // --------------------
    // then, handle special cases for locality
    special_rec_cases(M,maxscore, a0l, a1l, j0, j1, a0r, a1r, idxs);
    
    // --------------------
    // trace recursively
    
    trace_M_matrices(a0l,a1l,a0r,a1r,idxs[q].q,idxs[q].i0,idxs[q].i1);
    
    if  ( idxs[q].i0+1 == p0.adj(j0) && idxs[q].i1+1 ==  p1.adj(j1) ) {
      // arc-match, since maximal predecessor is one before arc in both seq/strs.
      trace_arc_match(p0.index(j0), p1.index(j1));
    } else if(idxs[q].q==q) { // only exclusions change q!
      int k0=-1, k1=-1;
      if (idxs[q].i0 == j0-1)
	k0=j0;
      if (idxs[q].i1 == j1-1)
	k1=j1;
      alignment.append(k0,k1);
    } else { // q has changed
      //cout << "exclusion from "<<idxs[q].i0<<","<<idxs[q].i1
      //    << " to "<<j0<<","<<j1<<endl;
    }
  } else { // j0==a0l || j1==a1l
    
    if (j0==a0l && j1==a1l) return;
    
    //cout << "termination handling in trace_M_matrices:"
    //      <<" "<<a0l<<" "<<a1l <<"; "<<j0<<" "<<j1<<endl;
        
    Index idxs[4];
    score_t maxscore[4];
  
    // --------------------
    // init maxscore
    for (int q_=0; q_<4; q_++)
      maxscore[q_]=NEG_SCORE_INFTY;
    
    // --------------------
    // first, handle the cases that are equal for all submatrices
    for (int q_=0; q_<4; q_++) {	    
      if (j0>a0l) {
	maxscore[q_] = std_init_column(M[q],j0,j1);
	idxs[q_].i0 = j0-1;
	idxs[q_].i1 = j1;
      } else {
	maxscore[q_] = std_init_row(M[q],j0,j1);
	idxs[q_].i0 = j0;
	idxs[q_].i1 = j1-1;
      }
      idxs[q_].q=q_;
    }
    
    // --------------------
    // then, handle special cases for locality
    special_rec_cases(M,maxscore, a0l, a1l, j0, j1, a0r, a1r, idxs);
    
    // --------------------
    // trace recursively
    
    trace_M_matrices(a0l,a1l,a0r,a1r,idxs[q].q,idxs[q].i0,idxs[q].i1);
        
    if (idxs[q].q == q) {
      // gap
      if (j0>a0l)
	alignment.append(j0,-1);
      else
	alignment.append(-1,j1);
    } else {
      //exclusion
    }

    /*
      Earlier quick and wrong assumptions, now replaced/corrected by above:
      if ( M[q].elem(j0,j1)==0 ) {
      // cout <<"terminal exclusion from "<<a0l<<","<<a1l<<" to "<<j0<<","<<j1 <<endl;
      } else {
      //cout <<"ordinary terminal gap"<<" M["<<j0<<","<<j1<<"]="<<M[q].elem(j0,j1)<<endl;
      for (int k0=a0l+1; k0<=j0; k0++)
      alignment.append(k0,-1);
      for (int k1=a1l+1; k1<=j1; k1++)
	alignment.append(-1,k1);
    }
    */
  }
}

void LSSAligner::trace_arc_match(int a0, int a1) {
  int a0l = p0[a0].first;
  int a1l = p1[a1].first;
  
  int a0r = p0[a0].second;
  int a1r = p1[a1].second;
  
  //cout << "trace arc match "<<a0<<" "<<a1
  //     <<"; "<<a0l<<".."<<a0r<<"; "<<a1l<<".."<<a1r<<endl;
  
  // recompute matrices m
  compute_M_matrices(a0l,a1l,a0r,a1r);

  alignment.append( p0[a0].first, p1[a1].first );
  trace_M_matrices(a0l,a1l,a0r,a1r,0,a0r-1,a1r-1);
  alignment.append( p0[a0].second, p1[a1].second );
}

// i0,i1 are the left ends of the aligned sequences (not -1)
void LSSAligner::trace_top_level(int i0, int i1, int j0, int j1) {
  if (j0<i0 || j1<i1 || (T.elem(j0,j1)==0 && p0.arc_compl(i0,j0) && p1.arc_compl(i1,j1))) {
    // ready
  } else {
    Index idx;
    std_rec_cases(NEG_SCORE_INFTY, T, i0-1, i1-1, j0, j1, slen0+1, slen1+1, &idx);
    trace_top_level(i0,i1,idx.i0,idx.i1);
    
    // test if we have an arc-match
    if ( idx.i0+1 == p0.adj(j0) && idx.i1+1 ==  p1.adj(j1) ) {
      trace_arc_match(p0.index(j0), p1.index(j1));
    } else {
      int k0=-1, k1=-1;
      if (idx.i0 == j0-1)
	k0=j0;
      if (idx.i1 == j1-1)
	k1=j1;
      alignment.append(k0,k1);
    }
  }
}

alignment_t LSSAligner::traceback() {
  compute_top_level(max_l0,max_l1,max_i0+1,max_i1+1);
  if (VERBOSE)
    T.print(cout,max_l0,max_l1,max_i0+1,max_i1+1);
  trace_top_level(max_l0+1,max_l1+1,max_i0,max_i1);
  
  /*
  cout <<"arc-complete(" << (max_l0+1) << ", " << max_i0 << ") = "
       <<p0.arc_compl(max_l0+1,max_i0)<<endl
       <<"arc-complete(" << (max_l1+1) << ", " << max_i1 << ") = "
       <<p1.arc_compl(max_l1+1,max_i1)<<endl;

  cout <<p1.left_vec[max_i1]<<", "<<p1.right_vec[max_i1];
  */

  return alignment;
}


// ===========================================================================
//
// Printing of trace
//
// ===========================================================================
//
//   print_alignment(): human readable output
//   print_tcoffee():   output prepared for t-coffee
//
// ===========================================================================


void LSSAligner::print_alignment() {
  int i0=0, i1=0;
  string a0,a1,a2,b0,b1;

  if (VERBOSE)
    for (Alignment::const_iterator i=alignment.begin(); 
	 i != alignment.end(); ++i) 
      cout << i->first <<"," <<i->second <<endl;
  
  for (Alignment::const_iterator i=alignment.begin(); 
       i != alignment.end(); ++i) {
    if ( i->first != -1 )
      for (int k=i0+1; k<i->first; k++) {
	a0 += s0[k];
	b0 += p0.toChar(k);
	a2 += ' ';
	a1 += ' ';
	b1 += ' ';
      }
    if ( i->second != -1 )
      for (int k=i1+1; k<i->second; k++) {
	a0 += ' ';
	b0 += ' ';
	a2 += ' ';
	a1 += s1[k];
	b1 += p1.toChar(k);
      }
    
    if ( i->first == -1 ) {
      a0 += '-';
      b0 += ' ';
    }
    else {
      i0 = i->first;
      a0 += s0[i0];
      b0 += p0.toChar(i0);
    }
    if ( i->second == -1 ) {
      a1 += '-';
      b1 += ' ';
    }
    else {
      i1 = i->second;
      a1 += s1[i1];
      b1 += p1.toChar(i1);
    }

    a2 += '|';
    
  }
  for (int k=i0+1; k<=s0.length(); k++) {
    a0 += s0[k];
    b0 += p0.toChar(k);
    a2 += ' ';
    a1 += ' ';
    b1 += ' ';
  }
  
  for (int k=i1+1; k<=s1.length(); k++) {
    a0 += ' ';
    b0 += ' ';
    a2 += ' ';
    a1 += s1[k];
    b1 += p1.toChar(k);
  }


  // pretty print the alignmentstrings
  {
    int line=83;
    vector<string> str;
    str.push_back(b0);
    str.push_back(a0);
    str.push_back(a2);
    str.push_back(a1);
    str.push_back(b1);
    
    size_t len = str[0].length();
    for (uint k=1; k<str.size(); k++)
      len = max(len,str[k].length());
    
    for (uint i=0; i<len; ) {
      uint j=i;
      for (uint k=0; k<str.size(); k++) {
	//cout << i << ": ";
	j=i;
	while (j<len) {
	  if (j < str[k].length()) cout << str[k][j];
	  j++;
	  if ((j)%line==0 ) break;  
	}
	cout <<endl;
      }
      i=j;
      cout <<endl;
    }
  }
}


bool
LSSAligner::is_edge_of_arcmatch(int i0, int i1) const {
  // pre: i0,i1 is an edge
  return p0.inc(i0) && p1.inc(i1)
    && alignment.is_edge(p0.adj(i0),p1.adj(i1));
}

bool
LSSAligner::is_edge_of_arcbreak(int k, int i0, int i1) const {
  // pre: i0,i1 is an edge
  if (k==0) {
    return p0.inc(i0) && 
      (!p1.inc(i1) || !alignment.is_edge(p0.adj(i0),p1.adj(i1)));
  }
  else {
    return p1.inc(i1) && 
      (!p0.inc(i0) || !alignment.is_edge(p0.adj(i0),p1.adj(i1)));
  }
}


void LSSAligner::print_tcoffee(score_t score, 
			       bool different_edge_weights=false) const{
  int i0=0, i1=0;
  
  for (Alignment::const_iterator i=alignment.begin();
       i != alignment.end(); ++i) {    
    
    if ( i->first != -1 && i->second!=-1 ) {
      score_t edgescore;
      i0=i->first;
      i1=i->second;
      
      edgescore = score;

      if (different_edge_weights) {
	if ( is_edge_of_arcmatch(i0,i1) )
	  edgescore += 10;
	
	if ( is_edge_of_arcbreak(0,i0,i1) )
	  edgescore -= 4;
	
	if ( is_edge_of_arcbreak(1,i0,i1) )
	  edgescore -= 4;
	
	if ( s0[i0]==s1[i1] ) {
	  edgescore += 5;
	}
      }
      
      cout << i0<<" "<<i1<<" "<<(int)edgescore<<" 1 0"<<endl;
    }
  }
}
